// @HEADER
// ************************************************************************
//
//               Rapid Optimization Library (ROL) Package
//                 Copyright (2014) Sandia Corporation
//
// Under terms of Contract DE-AC04-94AL85000, there is a non-exclusive
// license for use of this work by or on behalf of the U.S. Government.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
// 1. Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
//
// 2. Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
//
// 3. Neither the name of the Corporation nor the names of the
// contributors may be used to endorse or promote products derived from
// this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY SANDIA CORPORATION "AS IS" AND ANY
// EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL SANDIA CORPORATION OR THE
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// Questions? Contact lead developers:
//              Drew Kouri   (dpkouri@sandia.gov) and
//              Denis Ridzal (dridzal@sandia.gov)
//
// ************************************************************************
// @HEADER

/*! \file  pde.hpp
    \brief Implements the local PDE interface for the fractional Poisson
           control problem.
*/

#ifndef PDE_FRACTIONAL_POISSON_HPP
#define PDE_FRACTIONAL_POISSON_HPP

#include "../../TOOLS/pde.hpp"
#include "../../TOOLS/fe.hpp"

#include "Intrepid_HGRAD_QUAD_C1_FEM.hpp"
#include "Intrepid_HGRAD_QUAD_C2_FEM.hpp"
#include "Intrepid_HGRAD_LINE_C1_FEM.hpp"
#include "Intrepid_HGRAD_LINE_Cn_FEM.hpp"
#include "Intrepid_DefaultCubatureFactory.hpp"
#include "Intrepid_FunctionSpaceTools.hpp"
#include "Intrepid_CellTools.hpp"

#include "ROL_SharedPointer.hpp"

template <class Real>
class PDE_Fractional_Poisson_Local : public PDE<Real> {
private:
  // Finite element basis information
  ROL::SharedPointer<Intrepid::Basis<Real, Intrepid::FieldContainer<Real> > > basisPtr_;
  std::vector<ROL::SharedPointer<Intrepid::Basis<Real, Intrepid::FieldContainer<Real> > > > basisPtrs_;
  // Cell cubature information
  ROL::SharedPointer<Intrepid::Cubature<Real> > cellCub_;
  // Cell node information
  ROL::SharedPointer<Intrepid::FieldContainer<Real> > volCellNodes_;
  std::vector<std::vector<ROL::SharedPointer<Intrepid::FieldContainer<Real> > > > bdryCellNodes_;
  std::vector<std::vector<std::vector<int> > > bdryCellLocIds_;
  // Finite element definition
  ROL::SharedPointer<FE<Real> > fe_vol_;
  // Local degrees of freedom on boundary, for each side of the reference cell (first index).
  std::vector<std::vector<int> > fidx_;
  // Coordinates of degrees freedom on boundary cells.
  // Indexing:  [sideset number][local side id](cell number, value at dof)
  std::vector<std::vector<ROL::SharedPointer<Intrepid::FieldContainer<Real> > > > bdryCellDofValues_;
  // Diffusivity Coefficient
  ROL::SharedPointer<Intrepid::FieldContainer<Real> > kappa_;
  // Right hand side
  ROL::SharedPointer<Intrepid::FieldContainer<Real> > rhs_;

  Real fracPower_;
  Real alpha_;

public:
  PDE_Fractional_Poisson_Local(Teuchos::ParameterList &parlist) {
    // Finite element fields.
    int basisOrder = parlist.sublist("Problem").get("Diffusion FE Basis Order",1);
    if (basisOrder == 1) {
      basisPtr_ = ROL::makeShared<Intrepid::Basis_HGRAD_QUAD_C1_FEM<Real, Intrepid::FieldContainer<Real> >>();
    }
    else if (basisOrder == 2) {
      basisPtr_ = ROL::makeShared<Intrepid::Basis_HGRAD_QUAD_C2_FEM<Real, Intrepid::FieldContainer<Real> >>();
    }
    basisPtrs_.clear(); basisPtrs_.push_back(basisPtr_);
    // Quadrature rules.
    shards::CellTopology cellType = basisPtr_->getBaseCellTopology();              // get the cell type from any basis
    Intrepid::DefaultCubatureFactory<Real> cubFactory;                             // create cubature factory
    int cubDegree = parlist.sublist("Problem").get("Diffusion Cubature Degree",2); // set cubature degree, e.g., 2
    cellCub_ = cubFactory.create(cellType, cubDegree);                             // create default cubature
    // Coefficients
    fracPower_ = parlist.sublist("Problem").get("Fractional Power",0.5);
    alpha_     = static_cast<Real>(1) - static_cast<Real>(2)*fracPower_;
  }

  void residual(ROL::SharedPointer<Intrepid::FieldContainer<Real> > & res,
                const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & u_coeff,
                const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & z_coeff = ROL::nullPointer,
                const ROL::SharedPointer<const std::vector<Real> > & z_param = ROL::nullPointer) {
    // GET DIMENSIONS
    const int c = fe_vol_->gradN()->dimension(0);
    const int f = fe_vol_->gradN()->dimension(1);
    const int p = fe_vol_->gradN()->dimension(2);
    const int d = fe_vol_->gradN()->dimension(3);
    // INITIALIZE RESIDUAL
    res = ROL::makeShared<Intrepid::FieldContainer<Real>>(c, f);
    // COMPUTE STIFFNESS TERM
    ROL::SharedPointer<Intrepid::FieldContainer<Real> > gradU_eval
      = ROL::makeShared<Intrepid::FieldContainer<Real>>(c, p, d);
    fe_vol_->evaluateGradient(gradU_eval, u_coeff);
    ROL::SharedPointer<Intrepid::FieldContainer<Real> > kappa_gradU_eval
      = ROL::makeShared<Intrepid::FieldContainer<Real>>(c, p, d);
    Intrepid::FunctionSpaceTools::tensorMultiplyDataData<Real>(*kappa_gradU_eval,*kappa_,*gradU_eval);
    Intrepid::FunctionSpaceTools::integrate<Real>(*res,
                                                  *kappa_gradU_eval,
                                                  *(fe_vol_->gradNdetJ()),
                                                  Intrepid::COMP_CPP,
                                                  false);
    // ADD RHS TO RESIDUAL
    Intrepid::FunctionSpaceTools::integrate<Real>(*res,
                                                  *rhs_,
                                                  *(fe_vol_->NdetJ()),
                                                  Intrepid::COMP_CPP,
                                                  true);
    // ADD CONTROL TERM TO RESIDUAL
    ROL::SharedPointer<Intrepid::FieldContainer<Real> > valZ_eval
      = ROL::makeShared<Intrepid::FieldContainer<Real>>(c, p);
    fe_vol_->evaluateValue(valZ_eval, z_coeff);
    Intrepid::RealSpaceTools<Real>::scale(*valZ_eval,static_cast<Real>(-1));
    Intrepid::FunctionSpaceTools::integrate<Real>(*res,
                                                  *valZ_eval,
                                                  *(fe_vol_->NdetJ()),
                                                  Intrepid::COMP_CPP,
                                                  true);
    // APPLY DIRICHLET CONDITIONS
    int numSideSets = bdryCellLocIds_.size();
    if (numSideSets > 0) {
      for (int i = 0; i < numSideSets; ++i) {
        int numLocalSideIds = bdryCellLocIds_[i].size();
        for (int j = 0; j < numLocalSideIds; ++j) {
          int numCellsSide = bdryCellLocIds_[i][j].size();
          int numBdryDofs = fidx_[j].size();
          for (int k = 0; k < numCellsSide; ++k) {
            int cidx = bdryCellLocIds_[i][j][k];
            for (int l = 0; l < numBdryDofs; ++l) {
              (*res)(cidx,fidx_[j][l]) = (*u_coeff)(cidx,fidx_[j][l]) - (*bdryCellDofValues_[i][j])(k,fidx_[j][l]);
            }
          }
        }
      }
    }
  }

  void Jacobian_1(ROL::SharedPointer<Intrepid::FieldContainer<Real> > & jac,
                  const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & u_coeff,
                  const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & z_coeff = ROL::nullPointer,
                  const ROL::SharedPointer<const std::vector<Real> > & z_param = ROL::nullPointer) {
    // GET DIMENSIONS
    const int c = fe_vol_->gradN()->dimension(0);
    const int f = fe_vol_->gradN()->dimension(1);
    const int p = fe_vol_->gradN()->dimension(2);
    const int d = fe_vol_->gradN()->dimension(3);
    // INITILAIZE JACOBIAN
    jac = ROL::makeShared<Intrepid::FieldContainer<Real>>(c, f, f);
    // COMPUTE STIFFNESS TERM
    ROL::SharedPointer<Intrepid::FieldContainer<Real> > kappa_gradN_eval
      = ROL::makeShared<Intrepid::FieldContainer<Real>>(c, f, p, d);
    Intrepid::FunctionSpaceTools::tensorMultiplyDataField<Real>(*kappa_gradN_eval,*kappa_,*(fe_vol_->gradN()));
    Intrepid::FunctionSpaceTools::integrate<Real>(*jac,
                                                  *kappa_gradN_eval,
                                                  *(fe_vol_->gradNdetJ()),
                                                  Intrepid::COMP_CPP,
                                                  false);
    // APPLY DIRICHLET CONDITIONS
    int numSideSets = bdryCellLocIds_.size();
    if (numSideSets > 0) {
      for (int i = 0; i < numSideSets; ++i) {
        int numLocalSideIds = bdryCellLocIds_[i].size();
        for (int j = 0; j < numLocalSideIds; ++j) {
          int numCellsSide = bdryCellLocIds_[i][j].size();
          int numBdryDofs = fidx_[j].size();
          for (int k = 0; k < numCellsSide; ++k) {
            int cidx = bdryCellLocIds_[i][j][k];
            for (int l = 0; l < numBdryDofs; ++l) {
              for (int m = 0; m < f; ++m) {
                (*jac)(cidx,fidx_[j][l],m) = static_cast<Real>(0);
              }
              (*jac)(cidx,fidx_[j][l],fidx_[j][l]) = static_cast<Real>(1);
            }
          }
        }
      }
    }
  }

  void Jacobian_2(ROL::SharedPointer<Intrepid::FieldContainer<Real> > & jac,
                const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & u_coeff,
                const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & z_coeff = ROL::nullPointer,
                const ROL::SharedPointer<const std::vector<Real> > & z_param = ROL::nullPointer) {
    // GET DIMENSIONS
    const int c = fe_vol_->gradN()->dimension(0);
    const int f = fe_vol_->gradN()->dimension(1);
    // INITIALIZE JACOBIAN
    jac = ROL::makeShared<Intrepid::FieldContainer<Real>>(c, f, f);
    // ADD CONTROL TERM
    *jac = *fe_vol_->massMat();
    Intrepid::RealSpaceTools<Real>::scale(*jac, static_cast<Real>(-1));
    // APPLY DIRICHLET CONDITIONS
    int numSideSets = bdryCellLocIds_.size();
    if (numSideSets > 0) {
      for (int i = 0; i < numSideSets; ++i) {
        int numLocalSideIds = bdryCellLocIds_[i].size();
        for (int j = 0; j < numLocalSideIds; ++j) {
          int numCellsSide = bdryCellLocIds_[i][j].size();
          int numBdryDofs = fidx_[j].size();
          for (int k = 0; k < numCellsSide; ++k) {
            int cidx = bdryCellLocIds_[i][j][k];
            for (int l = 0; l < numBdryDofs; ++l) {
              for (int m = 0; m < f; ++m) {
                (*jac)(cidx,fidx_[j][l],m) = static_cast<Real>(0);
              }
            }
          }
        }
      }
    }
  }

  void Hessian_11(ROL::SharedPointer<Intrepid::FieldContainer<Real> > & hess,
                  const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & l_coeff,
                  const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & u_coeff,
                  const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & z_coeff = ROL::nullPointer,
                  const ROL::SharedPointer<const std::vector<Real> > & z_param = ROL::nullPointer) {
    throw Exception::Zero(">>> (PDE_Poisson::Hessian_11): Hessian is zero.");
  }

  void Hessian_12(ROL::SharedPointer<Intrepid::FieldContainer<Real> > & hess,
                  const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & l_coeff,
                  const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & u_coeff,
                  const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & z_coeff = ROL::nullPointer,
                  const ROL::SharedPointer<const std::vector<Real> > & z_param = ROL::nullPointer) {
    throw Exception::Zero(">>> (PDE_Poisson::Hessian_12): Hessian is zero.");
  }

  void Hessian_21(ROL::SharedPointer<Intrepid::FieldContainer<Real> > & hess,
                  const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & l_coeff,
                  const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & u_coeff,
                  const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & z_coeff = ROL::nullPointer,
                  const ROL::SharedPointer<const std::vector<Real> > & z_param = ROL::nullPointer) {
    throw Exception::Zero(">>> (PDE_Poisson::Hessian_21): Hessian is zero.");
  }

  void Hessian_22(ROL::SharedPointer<Intrepid::FieldContainer<Real> > & hess,
                  const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & l_coeff,
                  const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & u_coeff,
                  const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & z_coeff = ROL::nullPointer,
                  const ROL::SharedPointer<const std::vector<Real> > & z_param = ROL::nullPointer) {
    throw Exception::Zero(">>> (PDE_Poisson::Hessian_22): Hessian is zero.");
  }

  void RieszMap_1(ROL::SharedPointer<Intrepid::FieldContainer<Real> > & riesz) {
    // GET DIMENSIONS
    const int c = fe_vol_->gradN()->dimension(0);
    const int f = fe_vol_->gradN()->dimension(1);
    // INITIALIZE RIESZ MAP
    riesz = ROL::makeShared<Intrepid::FieldContainer<Real>>(c, f, f);
    *riesz = *fe_vol_->massMat();
    // APPLY DIRICHLET CONDITIONS
    int numSideSets = bdryCellLocIds_.size();
    if (numSideSets > 0) {
      for (int i = 0; i < numSideSets; ++i) {
        int numLocalSideIds = bdryCellLocIds_[i].size();
        for (int j = 0; j < numLocalSideIds; ++j) {
          int numCellsSide = bdryCellLocIds_[i][j].size();
          int numBdryDofs = fidx_[j].size();
          for (int k = 0; k < numCellsSide; ++k) {
            int cidx = bdryCellLocIds_[i][j][k];
            for (int l = 0; l < numBdryDofs; ++l) {
              for (int m = 0; m < f; ++m) {
                (*riesz)(cidx,fidx_[j][l],m) = static_cast<Real>(0);
              }
              (*riesz)(cidx,fidx_[j][l],fidx_[j][l]) = static_cast<Real>(1);
            }
          }
        }
      }
    }
  }

  void RieszMap_2(ROL::SharedPointer<Intrepid::FieldContainer<Real> > & riesz) {
    // GET DIMENSIONS
    const int c = fe_vol_->gradN()->dimension(0);
    const int f = fe_vol_->gradN()->dimension(1);
    // INITIALIZE RIESZ MAP
    riesz = ROL::makeShared<Intrepid::FieldContainer<Real>>(c, f, f);
    *riesz = *fe_vol_->massMat();
  }

  std::vector<ROL::SharedPointer<Intrepid::Basis<Real, Intrepid::FieldContainer<Real> > > > getFields() {
    return basisPtrs_;
  }

  void setCellNodes(const ROL::SharedPointer<Intrepid::FieldContainer<Real> > &volCellNodes,
                    const std::vector<std::vector<ROL::SharedPointer<Intrepid::FieldContainer<Real> > > > &bdryCellNodes,
                    const std::vector<std::vector<std::vector<int> > > &bdryCellLocIds) {
    volCellNodes_ = volCellNodes;
    bdryCellNodes_ = bdryCellNodes;
    bdryCellLocIds_ = bdryCellLocIds;
    // Finite element definition.
    fe_vol_ = ROL::makeShared<FE<Real>>(volCellNodes_,basisPtr_,cellCub_);
    // Set local boundary DOFs.
    fidx_ = fe_vol_->getBoundaryDofs();
    // Compute Dirichlet values at DOFs.
    int d = basisPtr_->getBaseCellTopology().getDimension();
    int numSidesets = bdryCellLocIds_.size();
    bdryCellDofValues_.resize(numSidesets);
    for (int i=0; i<numSidesets; ++i) {
      int numLocSides = bdryCellLocIds_[i].size();
      bdryCellDofValues_[i].resize(numLocSides);
      for (int j=0; j<numLocSides; ++j) {
        int c = bdryCellLocIds_[i][j].size();
        int f = basisPtr_->getCardinality();
        bdryCellDofValues_[i][j] = ROL::makeShared<Intrepid::FieldContainer<Real>>(c, f);
        ROL::SharedPointer<Intrepid::FieldContainer<Real> > coords =
          ROL::makeShared<Intrepid::FieldContainer<Real>>(c, f, d);
        if (c > 0) {
          fe_vol_->computeDofCoords(coords, bdryCellNodes_[i][j]);
        }
        for (int k=0; k<c; ++k) {
          for (int l=0; l<f; ++l) {
            std::vector<Real> dofpoint(d);
            for (int m=0; m<d; ++m) {
              dofpoint[m] = (*coords)(k, l, m);
            }
            (*bdryCellDofValues_[i][j])(k, l) = DirichletFunc(dofpoint, i, j);
          }
        }
      }
    }
    // Compute diffusivity coefficient
    const int c = fe_vol_->gradN()->dimension(0);
    const int p = fe_vol_->gradN()->dimension(2);
    kappa_ = ROL::makeShared<Intrepid::FieldContainer<Real>>(c,p);
    rhs_   = ROL::makeShared<Intrepid::FieldContainer<Real>>(c,p);
    computeCoefficients(kappa_,rhs_); 
  }

  const ROL::SharedPointer<FE<Real> > getFE(void) const {
    return fe_vol_;
  }

private:
  Real diffusivityFunc(const std::vector<Real> &x) const {
    return static_cast<Real>(1);
  }

  Real rhsFunc(const std::vector<Real> &x) const {
    const Real lambda = M_PI*M_PI*static_cast<Real>(8);
    const Real twoPi  = static_cast<Real>(2) * M_PI;
    return std::pow(lambda,fracPower_)*std::sin(twoPi*x[0])*std::sin(twoPi*x[1]);
  }

  Real DirichletFunc(const std::vector<Real> & coords, int sideset, int locSideId) const {
    return 0;
  }

  void computeCoefficients(const ROL::SharedPointer<Intrepid::FieldContainer<Real> > &kappa,
                           const ROL::SharedPointer<Intrepid::FieldContainer<Real> > &rhs) const {
    // GET DIMENSIONS
    int c = fe_vol_->gradN()->dimension(0);
    int p = fe_vol_->gradN()->dimension(2);
    int d = fe_vol_->gradN()->dimension(3);
    std::vector<Real> pt(d), adv(d);
    for (int i = 0; i < c; ++i) {
      for (int j = 0; j < p; ++j) {
        for ( int k = 0; k < d; ++k) {
          pt[k] = (*fe_vol_->cubPts())(i,j,k);
        }
        // Compute diffusivity kappa
        (*kappa)(i,j) = diffusivityFunc(pt);
        // Compute right hand side
        (*rhs)(i,j) = -rhsFunc(pt);
      }
    }
  }

}; // PDE_Fractional_Poisson_Local

template <class Real>
class PDE_Fractional_Poisson_Cylinder : public PDE<Real> {
private:
  // Finite element basis information
  ROL::SharedPointer<Intrepid::Basis<Real, Intrepid::FieldContainer<Real> > > basisPtr_;
  std::vector<ROL::SharedPointer<Intrepid::Basis<Real, Intrepid::FieldContainer<Real> > > > basisPtrs_;
  // Cell cubature information
  ROL::SharedPointer<Intrepid::Cubature<Real> > cellCub_;
  // Cell node information
  ROL::SharedPointer<Intrepid::FieldContainer<Real> > volCellNodes_;
  std::vector<std::vector<ROL::SharedPointer<Intrepid::FieldContainer<Real> > > > bdryCellNodes_;
  std::vector<std::vector<std::vector<int> > > bdryCellLocIds_;
  // Finite element definition
  ROL::SharedPointer<FE<Real> > fe_vol_;
  // Local degrees of freedom on boundary, for each side of the reference cell (first index).
  std::vector<std::vector<int> > fidx_;
  // Coordinates of degrees freedom on boundary cells.
  // Indexing:  [sideset number][local side id](cell number, value at dof)
  std::vector<std::vector<ROL::SharedPointer<Intrepid::FieldContainer<Real> > > > bdryCellDofValues_;

  Real fracPower_;
  Real alpha_;

  Real diffusivityFunc(const std::vector<Real> &x) const {
    Real alpha = alpha_;
    std::vector<Real> param = PDE<Real>::getParameter();
    if (param.size()) {
      alpha = static_cast<Real>(1) - static_cast<Real>(2)*param[0];
    }
    return std::pow(x[0],alpha);
  }

  Real DirichletFunc(const std::vector<Real> & coords, int sideset, int locSideId) const {
    return 0;
  }

  void computeCoefficients(ROL::SharedPointer<Intrepid::FieldContainer<Real> > &kappa) const {
    // GET DIMENSIONS
    int c = fe_vol_->gradN()->dimension(0);
    int p = fe_vol_->gradN()->dimension(2);
    int d = fe_vol_->gradN()->dimension(3);
    std::vector<Real> pt(d), adv(d);
    for (int i = 0; i < c; ++i) {
      for (int j = 0; j < p; ++j) {
        for ( int k = 0; k < d; ++k) {
          pt[k] = (*fe_vol_->cubPts())(i,j,k);
        }
        // Compute diffusivity kappa
        (*kappa)(i,j) = diffusivityFunc(pt);
      }
    }
  }

public:
  PDE_Fractional_Poisson_Cylinder(Teuchos::ParameterList &parlist) {
    // Finite element fields.
    int basisOrder = parlist.sublist("Problem").get("Cylinder FE Basis Order",1);
    if (basisOrder == 1) {
      basisPtr_ = ROL::makeShared<Intrepid::Basis_HGRAD_LINE_C1_FEM<Real, Intrepid::FieldContainer<Real> >>();
    }
    else if (basisOrder >= 2) {
      basisPtr_ = ROL::makeShared<Intrepid::Basis_HGRAD_LINE_Cn_FEM<Real, Intrepid::FieldContainer<Real> >>(basisOrder,Intrepid::POINTTYPE_EQUISPACED);
    }
    basisPtrs_.clear(); basisPtrs_.push_back(basisPtr_);
    // Quadrature rules.
    shards::CellTopology cellType = basisPtr_->getBaseCellTopology();             // get the cell type from any basis
    Intrepid::DefaultCubatureFactory<Real> cubFactory;                            // create cubature factory
    int cubDegree = parlist.sublist("Problem").get("Cylinder Cubature Degree",2); // set cubature degree, e.g., 2
    cellCub_ = cubFactory.create(cellType, cubDegree);                            // create default cubature
    // Coefficients
    fracPower_ = parlist.sublist("Problem").get("Fractional Power",0.5);
    alpha_     = static_cast<Real>(1) - static_cast<Real>(2)*fracPower_;
  }

  void residual(ROL::SharedPointer<Intrepid::FieldContainer<Real> > & res,
                const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & u_coeff,
                const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & z_coeff = ROL::nullPointer,
                const ROL::SharedPointer<const std::vector<Real> > & z_param = ROL::nullPointer) {
    // GET DIMENSIONS
    const int c = fe_vol_->gradN()->dimension(0);
    const int f = fe_vol_->gradN()->dimension(1);
    const int p = fe_vol_->gradN()->dimension(2);
    const int d = fe_vol_->gradN()->dimension(3);
    // INITIALIZE RESIDUAL
    res = ROL::makeShared<Intrepid::FieldContainer<Real>>(c, f);
    // EVALUATE DIFFUSIVITY
    ROL::SharedPointer<Intrepid::FieldContainer<Real> > kappa
      = ROL::makeShared<Intrepid::FieldContainer<Real>>(c, p);
    computeCoefficients(kappa);
    // COMPUTE STIFFNESS TERM
    ROL::SharedPointer<Intrepid::FieldContainer<Real> > gradU_eval
      = ROL::makeShared<Intrepid::FieldContainer<Real>>(c, p, d);
    ROL::SharedPointer<Intrepid::FieldContainer<Real> > kappa_gradU_eval
      = ROL::makeShared<Intrepid::FieldContainer<Real>>(c, p, d);
    fe_vol_->evaluateGradient(gradU_eval, u_coeff);
    Intrepid::FunctionSpaceTools::tensorMultiplyDataData<Real>(*kappa_gradU_eval,*kappa,*gradU_eval);
    Intrepid::FunctionSpaceTools::integrate<Real>(*res,
                                                  *kappa_gradU_eval,
                                                  *(fe_vol_->gradNdetJ()),
                                                  Intrepid::COMP_CPP,
                                                  false);
    // ADD CONTROL TERM TO RESIDUAL
    ROL::SharedPointer<Intrepid::FieldContainer<Real> > valZ_eval
      = ROL::makeShared<Intrepid::FieldContainer<Real>>(c, p);
    fe_vol_->evaluateValue(valZ_eval, z_coeff);
    Intrepid::RealSpaceTools<Real>::scale(*valZ_eval,static_cast<Real>(-1));
    Intrepid::FunctionSpaceTools::integrate<Real>(*res,
                                                  *valZ_eval,
                                                  *(fe_vol_->NdetJ()),
                                                  Intrepid::COMP_CPP,
                                                  true);
    // APPLY DIRICHLET CONDITIONS
    int numSideSets = bdryCellLocIds_.size();
    if (numSideSets > 0) {
      for (int i = 0; i < numSideSets; ++i) {
        if (i==1) {
          int numLocalSideIds = bdryCellLocIds_[i].size();
          for (int j = 0; j < numLocalSideIds; ++j) {
            int numCellsSide = bdryCellLocIds_[i][j].size();
            int numBdryDofs = fidx_[j].size();
            for (int k = 0; k < numCellsSide; ++k) {
              int cidx = bdryCellLocIds_[i][j][k];
              for (int l = 0; l < numBdryDofs; ++l) {
                (*res)(cidx,fidx_[j][l]) = (*u_coeff)(cidx,fidx_[j][l]) - (*bdryCellDofValues_[i][j])(k,fidx_[j][l]);
              }
            }
          }
        }
      }
    }
  }

  void Jacobian_1(ROL::SharedPointer<Intrepid::FieldContainer<Real> > & jac,
                  const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & u_coeff,
                  const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & z_coeff = ROL::nullPointer,
                  const ROL::SharedPointer<const std::vector<Real> > & z_param = ROL::nullPointer) {
    // GET DIMENSIONS
    const int c = fe_vol_->gradN()->dimension(0);
    const int f = fe_vol_->gradN()->dimension(1);
    const int p = fe_vol_->gradN()->dimension(2);
    const int d = fe_vol_->gradN()->dimension(3);
    // INITILAIZE JACOBIAN
    jac = ROL::makeShared<Intrepid::FieldContainer<Real>>(c, f, f);
    // EVALUATE DIFFUSIVITY
    ROL::SharedPointer<Intrepid::FieldContainer<Real> > kappa
      = ROL::makeShared<Intrepid::FieldContainer<Real>>(c, p);
    computeCoefficients(kappa);
    // COMPUTE STIFFNESS TERM
    ROL::SharedPointer<Intrepid::FieldContainer<Real> > kappa_gradN_eval
      = ROL::makeShared<Intrepid::FieldContainer<Real>>(c, f, p, d);
    Intrepid::FunctionSpaceTools::tensorMultiplyDataField<Real>(*kappa_gradN_eval,*kappa,*(fe_vol_->gradN()));
    Intrepid::FunctionSpaceTools::integrate<Real>(*jac,
                                                  *kappa_gradN_eval,
                                                  *(fe_vol_->gradNdetJ()),
                                                  Intrepid::COMP_CPP,
                                                  false);
    // APPLY DIRICHLET CONDITIONS
    int numSideSets = bdryCellLocIds_.size();
    if (numSideSets > 0) {
      for (int i = 0; i < numSideSets; ++i) {
        if (i==1) {
          int numLocalSideIds = bdryCellLocIds_[i].size();
          for (int j = 0; j < numLocalSideIds; ++j) {
            int numCellsSide = bdryCellLocIds_[i][j].size();
            int numBdryDofs = fidx_[j].size();
            for (int k = 0; k < numCellsSide; ++k) {
              int cidx = bdryCellLocIds_[i][j][k];
              for (int l = 0; l < numBdryDofs; ++l) {
                for (int m = 0; m < f; ++m) {
                  (*jac)(cidx,fidx_[j][l],m) = static_cast<Real>(0);
                }
                (*jac)(cidx,fidx_[j][l],fidx_[j][l]) = static_cast<Real>(1);
              }
            }
          }
        }
      }
    }
  }

  void Jacobian_2(ROL::SharedPointer<Intrepid::FieldContainer<Real> > & jac,
                  const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & u_coeff,
                  const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & z_coeff = ROL::nullPointer,
                  const ROL::SharedPointer<const std::vector<Real> > & z_param = ROL::nullPointer) {
    // GET DIMENSIONS
    const int c = fe_vol_->gradN()->dimension(0);
    const int f = fe_vol_->gradN()->dimension(1);
    // INITIALIZE JACOBIAN
    jac = ROL::makeShared<Intrepid::FieldContainer<Real>>(c, f, f);
    // ADD CONTROL TERM
    *jac = *fe_vol_->massMat();
    Intrepid::RealSpaceTools<Real>::scale(*jac, static_cast<Real>(-1));
    // APPLY DIRICHLET CONDITIONS
    int numSideSets = bdryCellLocIds_.size();
    if (numSideSets > 0) {
      for (int i = 0; i < numSideSets; ++i) {
        int numLocalSideIds = bdryCellLocIds_[i].size();
        for (int j = 0; j < numLocalSideIds; ++j) {
          int numCellsSide = bdryCellLocIds_[i][j].size();
          int numBdryDofs = fidx_[j].size();
          for (int k = 0; k < numCellsSide; ++k) {
            int cidx = bdryCellLocIds_[i][j][k];
            for (int l = 0; l < numBdryDofs; ++l) {
              for (int m = 0; m < f; ++m) {
                (*jac)(cidx,fidx_[j][l],m) = static_cast<Real>(0);
              }
            }
          }
        }
      }
    }
  }

  void Hessian_11(ROL::SharedPointer<Intrepid::FieldContainer<Real> > & hess,
                  const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & l_coeff,
                  const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & u_coeff,
                  const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & z_coeff = ROL::nullPointer,
                  const ROL::SharedPointer<const std::vector<Real> > & z_param = ROL::nullPointer) {
    throw Exception::Zero(">>> (PDE_Fractional_Poisson_Cylinder::Hessian_11): Hessian is zero.");
  }

  void Hessian_12(ROL::SharedPointer<Intrepid::FieldContainer<Real> > & hess,
                  const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & l_coeff,
                  const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & u_coeff,
                  const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & z_coeff = ROL::nullPointer,
                  const ROL::SharedPointer<const std::vector<Real> > & z_param = ROL::nullPointer) {
    throw Exception::Zero(">>> (PDE_Fractional_Poisson_Cylinder::Hessian_12): Hessian is zero.");
  }

  void Hessian_21(ROL::SharedPointer<Intrepid::FieldContainer<Real> > & hess,
                  const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & l_coeff,
                  const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & u_coeff,
                  const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & z_coeff = ROL::nullPointer,
                  const ROL::SharedPointer<const std::vector<Real> > & z_param = ROL::nullPointer) {
    throw Exception::Zero(">>> (PDE_Fractional_Poisson_Cylinder::Hessian_21): Hessian is zero.");
  }

  void Hessian_22(ROL::SharedPointer<Intrepid::FieldContainer<Real> > & hess,
                  const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & l_coeff,
                  const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & u_coeff,
                  const ROL::SharedPointer<const Intrepid::FieldContainer<Real> > & z_coeff = ROL::nullPointer,
                  const ROL::SharedPointer<const std::vector<Real> > & z_param = ROL::nullPointer) {
    throw Exception::Zero(">>> (PDE_Fractional_Poisson_Cylinder::Hessian_22): Hessian is zero.");
  }

  void RieszMap_1(ROL::SharedPointer<Intrepid::FieldContainer<Real> > & riesz) {
    // GET DIMENSIONS
    const int c = fe_vol_->gradN()->dimension(0);
    const int f = fe_vol_->gradN()->dimension(1);
    const int p = fe_vol_->gradN()->dimension(2);
    // INITIALIZE RIESZ MAP
    riesz = ROL::makeShared<Intrepid::FieldContainer<Real>>(c, f, f);
    // EVALUATE DIFFUSIVITY
    ROL::SharedPointer<Intrepid::FieldContainer<Real> > kappa
      = ROL::makeShared<Intrepid::FieldContainer<Real>>(c, p);
    computeCoefficients(kappa);
    // COMPUTE RIESZ MAP
    ROL::SharedPointer<Intrepid::FieldContainer<Real> > kappa_N_eval
      = ROL::makeShared<Intrepid::FieldContainer<Real>>(c, f, p);
    Intrepid::FunctionSpaceTools::scalarMultiplyDataField<Real>(*kappa_N_eval,*kappa,*(fe_vol_->N()));
    Intrepid::FunctionSpaceTools::integrate<Real>(*riesz,
                                                  *kappa_N_eval,
                                                  *(fe_vol_->NdetJ()),
                                                  Intrepid::COMP_CPP, false);
    // APPLY DIRICHLET CONDITIONS
    int numSideSets = bdryCellLocIds_.size();
    if (numSideSets > 0) {
      for (int i = 0; i < numSideSets; ++i) {
        if (i==1) {
          int numLocalSideIds = bdryCellLocIds_[i].size();
          for (int j = 0; j < numLocalSideIds; ++j) {
            int numCellsSide = bdryCellLocIds_[i][j].size();
            int numBdryDofs = fidx_[j].size();
            for (int k = 0; k < numCellsSide; ++k) {
              int cidx = bdryCellLocIds_[i][j][k];
              for (int l = 0; l < numBdryDofs; ++l) {
                for (int m = 0; m < f; ++m) {
                  (*riesz)(cidx,fidx_[j][l],m) = static_cast<Real>(0);
                }
                (*riesz)(cidx,fidx_[j][l],fidx_[j][l]) = static_cast<Real>(1);
              }
            }
          }
        }
      }
    }
  }

  void RieszMap_2(ROL::SharedPointer<Intrepid::FieldContainer<Real> > & riesz) {
    // GET DIMENSIONS
    const int c = fe_vol_->gradN()->dimension(0);
    const int f = fe_vol_->gradN()->dimension(1);
    // INITIALIZE RIESZ MAP
    riesz = ROL::makeShared<Intrepid::FieldContainer<Real>>(c, f, f);
    *riesz = *fe_vol_->massMat();
  }

  std::vector<ROL::SharedPointer<Intrepid::Basis<Real, Intrepid::FieldContainer<Real> > > > getFields() {
    return basisPtrs_;
  }

  void setCellNodes(const ROL::SharedPointer<Intrepid::FieldContainer<Real> > &volCellNodes,
                    const std::vector<std::vector<ROL::SharedPointer<Intrepid::FieldContainer<Real> > > > &bdryCellNodes,
                    const std::vector<std::vector<std::vector<int> > > &bdryCellLocIds) {
    volCellNodes_ = volCellNodes;
    bdryCellNodes_ = bdryCellNodes;
    bdryCellLocIds_ = bdryCellLocIds;
    // Finite element definition.
    fe_vol_ = ROL::makeShared<FE<Real>>(volCellNodes_,basisPtr_,cellCub_);
    // Set local boundary DOFs.
    fidx_ = fe_vol_->getBoundaryDofs();
    // Compute Dirichlet values at DOFs.
    int d = basisPtr_->getBaseCellTopology().getDimension();
    int numSidesets = bdryCellLocIds_.size();
    bdryCellDofValues_.resize(numSidesets);
    for (int i=0; i<numSidesets; ++i) {
      int numLocSides = bdryCellLocIds_[i].size();
      bdryCellDofValues_[i].resize(numLocSides);
      for (int j=0; j<numLocSides; ++j) {
        int c = bdryCellLocIds_[i][j].size();
        int f = basisPtr_->getCardinality();
        bdryCellDofValues_[i][j] = ROL::makeShared<Intrepid::FieldContainer<Real>>(c, f);
        ROL::SharedPointer<Intrepid::FieldContainer<Real> > coords =
          ROL::makeShared<Intrepid::FieldContainer<Real>>(c, f, d);
        if (c > 0) {
          fe_vol_->computeDofCoords(coords, bdryCellNodes_[i][j]);
        }
        for (int k=0; k<c; ++k) {
          for (int l=0; l<f; ++l) {
            std::vector<Real> dofpoint(d);
            for (int m=0; m<d; ++m) {
              dofpoint[m] = (*coords)(k, l, m);
            }
            (*bdryCellDofValues_[i][j])(k, l) = DirichletFunc(dofpoint, i, j);
          }
        }
      }
    }
  }

  const ROL::SharedPointer<FE<Real> > getFE(void) const {
    return fe_vol_;
  }

}; // PDE_Fractional_Poisson_Local

#endif
